from requests import HTTPError


class OsoError(Exception):
    pass


class OsoHTTPError(HTTPError):
    def __str__(self):
        str = super().__str__()
        response = self.response.json()
        return f"{str} - {response['error'] if response['error'] else response}"


class InsufficientCreditError(OsoError):
    """Raised when the user has insufficient credits to execute a query."""



from typing import Callable, Optional

import pandas as pd
import requests
from pydantic import BaseModel, Field
from pyoso.exceptions import OsoHTTPError


class SemanticColumn(BaseModel):
    name: str
    type: str
    description: Optional[str]


class SemanticRelationship(BaseModel):
    source_column: str = Field(alias="sourceColumn")
    target_table: str = Field(alias="targetTable")
    target_column: str = Field(alias="targetColumn")


class SemanticTableResponse(BaseModel):
    name: str
    description: Optional[str]
    columns: list[SemanticColumn] = Field(default_factory=list)
    relationships: list[SemanticRelationship] = Field(default_factory=list)


def query_dynamic_models(base_url: str, api_key: str) -> list[SemanticTableResponse]:
    headers = {
        "Content-Type": "application/json",
    }
    headers["Authorization"] = f"Bearer {api_key}"
    try:
        response = requests.get(
            f"{base_url}connector",
            headers=headers,
        )
        response.raise_for_status()

        response_data = response.json()

        tables = [
            SemanticTableResponse.model_validate(table) for table in response_data
        ]
        return tables
    except requests.HTTPError as e:
        raise OsoHTTPError(e, response=e.response) from None


def create_registry(
    base_url: str, api_key: str, to_pandas_fn: Callable[[str], pd.DataFrame]
):
    from oso_semantic import Dimension, Model
    from oso_semantic import QueryBuilder as InnerQueryBuilder
    from oso_semantic import (
        Registry,
        Relationship,
        RelationshipType,
        register_oso_models,
    )

    class QueryBuilder(InnerQueryBuilder):
        def __init__(self, registry: Registry):
            super().__init__(registry)

        def to_pandas(self):
            sql = self.build()
            return to_pandas_fn(sql.sql(dialect="trino"))

    registry = Registry(QueryBuilder)

    register_oso_models(registry)

    tables = query_dynamic_models(base_url, api_key)
    for table in tables:
        model_name = table.name.split(".")[-1]
        registry.register(
            Model(
                name=model_name,
                description=table.description or "",
                table=table.name,
                dimensions=[
                    Dimension(
                        name=column.name,
                        column_name=column.name,
                        description=column.description or "",
                    )
                    for column in table.columns
                ],
                relationships=[
                    Relationship(
                        name=f"{relationship.source_column}->{relationship.target_table.split('.')[-1]}.{relationship.target_column}",
                        type=RelationshipType.MANY_TO_ONE,
                        source_foreign_key=relationship.source_column,
                        ref_model=relationship.target_table.split(".")[-1],
                        ref_key=relationship.target_column,
                    )
                    for relationship in table.relationships
                ],
            )
        )

    return registry



import os
import sys
from importlib import reload
from unittest import TestCase, mock

import requests
from oso_semantic import Registry
from pyoso.client import Client, ClientConfig, QueryData, QueryResponse
from pyoso.exceptions import OsoError, OsoHTTPError


class TestClient(TestCase):
    DEFAULT_API_KEY = "test_key"
    CUSTOM_API_KEY = "custom_key"

    @mock.patch.dict(os.environ, {"OSO_API_KEY": ""})
    def test_constructor_without_api_key(self):
        with self.assertRaises(OsoError):
            Client(api_key=None)

    @mock.patch("pyoso.client.create_registry")
    @mock.patch("requests.get")
    @mock.patch("requests.post")
    def test_to_pandas(
        self, mock_post: mock.Mock, mock_get: mock.Mock, mock_registry: mock.Mock
    ):
        mock_registry.return_value = Registry()

        # Setup POST response
        mock_post_response = mock.Mock()
        mock_post_response.json.return_value = {
            "id": "query_123",
            "status": "queued",
            "url": None,
        }
        mock_post_response.raise_for_status.return_value = None
        mock_post.return_value = mock_post_response

        # Setup GET responses
        def side_effect_get(url, headers=None, params=None, **kwargs):
            mock_resp = mock.Mock()
            mock_resp.raise_for_status.return_value = None

            if "async-sql" in url and params and params.get("id") == "query_123":
                mock_resp.json.return_value = {
                    "id": "query_123",
                    "status": "completed",
                    "url": "http://s3-bucket/result.csv",
                }
            elif url == "http://s3-bucket/result.csv":
                # Mock iter_lines for streaming response
                mock_resp.iter_lines.return_value = iter(['["column"]', '["test"]'])
            else:
                mock_resp.status_code = 404

            return mock_resp

        mock_get.side_effect = side_effect_get

        client = Client(
            api_key=self.CUSTOM_API_KEY,
            client_opts=ClientConfig(base_url="http://localhost:8000/api/v1"),
        )
        query = "SELECT * FROM test_table"
        df = client.to_pandas(query)

        mock_post.assert_called_once_with(
            "http://localhost:8000/api/v1/async-sql",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.CUSTOM_API_KEY}",
            },
            json={"query": query},
        )

        self.assertEqual(df.columns.tolist(), ["column"])
        self.assertEqual(df.values.tolist(), [["test"]])

    @mock.patch.dict(os.environ, {"OSO_API_KEY": DEFAULT_API_KEY})
    @mock.patch("pyoso.client.create_registry")
    @mock.patch("requests.get")
    @mock.patch("requests.post")
    def test_to_pandas_with_default_api_key(
        self, mock_post: mock.Mock, mock_get: mock.Mock, mock_registry: mock.Mock
    ):
        mock_registry.return_value = Registry()

        # Setup POST response
        mock_post_response = mock.Mock()
        mock_post_response.json.return_value = {
            "id": "query_def",
            "status": "completed",
            "url": "http://s3-bucket/result_def.csv",
        }
        mock_post_response.raise_for_status.return_value = None
        mock_post.return_value = mock_post_response

        # Setup GET response
        def side_effect_get(url, **kwargs):
            mock_resp = mock.Mock()
            mock_resp.raise_for_status.return_value = None
            if url == "http://s3-bucket/result_def.csv":
                mock_resp.iter_lines.return_value = iter(['["column"]', '["test"]'])
            return mock_resp

        mock_get.side_effect = side_effect_get

        client = Client()
        query = "SELECT * FROM test_table"
        df = client.to_pandas(query)

        mock_post.assert_called_once_with(
            "https://www.oso.xyz/api/v1/async-sql",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.DEFAULT_API_KEY}",
            },
            json={"query": query},
        )

        mock_get.assert_called_with("http://s3-bucket/result_def.csv", stream=True)
        self.assertEqual(df.columns.tolist(), ["column"])
        self.assertEqual(df.values.tolist(), [["test"]])

    @mock.patch("pyoso.client.create_registry")
    @mock.patch("requests.post")
    def test_to_pandas_http_error(self, mock_post: mock.Mock, mock_registry: mock.Mock):
        mock_registry.return_value = Registry()
        mock_response = mock.Mock()
        mock_response.status_code = 500
        http_error = requests.HTTPError("HTTP Error")
        http_error.response = mock_response
        mock_response.raise_for_status.side_effect = http_error
        mock_post.return_value = mock_response

        client = Client(api_key=self.CUSTOM_API_KEY)
        query = "SELECT * FROM test_table"

        with self.assertRaises(OsoHTTPError):
            client.to_pandas(query)

        mock_post.assert_called_once_with(
            "https://www.oso.xyz/api/v1/async-sql",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.CUSTOM_API_KEY}",
            },
            json={"query": query},
        )

    @mock.patch("requests.get")
    @mock.patch("requests.post")
    def test_semantic_select_to_pandas(self, mock_post: mock.Mock, mock_get: mock.Mock):
        # Setup Semantic Registry GET response
        connector_response_data = [
            {
                "name": "oso.users",
                "description": "User information table",
                "columns": [
                    {"name": "id", "type": "bigint", "description": "User ID"},
                    {"name": "name", "type": "varchar", "description": "User name"},
                ],
                "relationships": [],
            }
        ]

        # Setup Async SQL responses
        mock_post_response = mock.Mock()
        mock_post_response.json.return_value = {
            "id": "query_sem",
            "status": "completed",
            "url": "http://s3/sem.csv",
        }
        mock_post_response.raise_for_status.return_value = None
        mock_post.return_value = mock_post_response

        def side_effect_get(url, headers=None, **kwargs):
            mock_resp = mock.Mock()
            mock_resp.raise_for_status.return_value = None

            if "connector" in url:
                mock_resp.json.return_value = connector_response_data
            elif url == "http://s3/sem.csv":
                mock_resp.iter_lines.return_value = iter(
                    ['["id", "name"]', '[1, "Alice"]', '[2, "Bob"]']
                )
            return mock_resp

        mock_get.side_effect = side_effect_get

        client = Client(api_key=self.CUSTOM_API_KEY)

        # Test the semantic select functionality
        query_builder = client.semantic.select("users.id", "users.name")
        result_df = query_builder.to_pandas()

        # Verify SQL execution
        self.assertTrue(mock_post.called)
        args, kwargs = mock_post.call_args
        self.assertEqual(args[0], "https://www.oso.xyz/api/v1/async-sql")

        # Check columns
        self.assertEqual(result_df.columns.tolist(), ["id", "name"])
        # Check data
        self.assertEqual(result_df.iloc[0]["name"], "Alice")
        self.assertEqual(str(result_df.iloc[0]["id"]), "1")

    @mock.patch("pyoso.client.create_registry")
    @mock.patch("requests.get")
    @mock.patch("requests.post")
    def test_query_with_analytics(
        self, mock_post: mock.Mock, mock_get: mock.Mock, mock_registry: mock.Mock
    ):
        mock_registry.return_value = Registry()

        # Setup POST response
        mock_post_response = mock.Mock()
        mock_post_response.json.return_value = {
            "id": "query_analytics",
            "status": "completed",
            "url": "http://s3/data.csv",
        }
        mock_post_response.raise_for_status.return_value = None
        mock_post.return_value = mock_post_response

        # Setup GET response
        def side_effect_get(url, **kwargs):
            mock_resp = mock.Mock()
            mock_resp.raise_for_status.return_value = None
            if url == "http://s3/data.csv":
                mock_resp.iter_lines.return_value = iter(
                    ['["column1", "column2"]', '["value1", "value2"]']
                )
            return mock_resp

        mock_get.side_effect = side_effect_get

        client = Client(api_key=self.CUSTOM_API_KEY)
        query = "SELECT * FROM oso.events"
        response = client.query(query)

        # Verify the response structure
        self.assertIsInstance(response, QueryResponse)
        self.assertIsInstance(response.data, QueryData)
        self.assertEqual(response.data.columns, ["column1", "column2"])
        self.assertEqual(response.data.data, [["value1", "value2"]])

        # In new implementation analytics is empty
        analytics = response.analytics
        self.assertIsInstance(analytics, (dict, object))  # DataAnalytics or dict
        # Verify it's effectively empty or has no keys
        self.assertEqual(len(analytics), 0)


@mock.patch.dict(sys.modules, {"oso_semantic": None})
class TestClientWithoutSemantic(TestCase):
    DEFAULT_API_KEY = "test_key"
    CUSTOM_API_KEY = "custom_key"

    @mock.patch.dict(sys.modules, {"oso_semantic": None})
    def test_constructor_without_semantic(self):
        reload(sys.modules["pyoso.client"])

        from pyoso.client import Client

        client = Client(api_key=self.CUSTOM_API_KEY)
        self.assertFalse(
            hasattr(client, "semantic"),
            "Semantic should not be initialized without oso_semantic",
        )



import json
import os
import time
from functools import lru_cache
from typing import Any, Literal, Optional

import pandas as pd
import requests
from pydantic import BaseModel, Field
from pyoso.analytics import DataAnalytics
from pyoso.constants import DEFAULT_BASE_URL, OSO_API_KEY
from pyoso.exceptions import InsufficientCreditError, OsoError, OsoHTTPError
from pyoso.semantic import create_registry
from sqlglot import parse

HAS_OSO_SEMANTIC = False
try:
    import oso_semantic  # noqa: F401

    HAS_OSO_SEMANTIC = True
except ImportError:
    pass


class ClientConfig(BaseModel):
    base_url: Optional[str]


class QueryData(BaseModel):
    columns: list[str]
    data: list[list[Any]]


class QueryResponse:
    """Response object containing query data and optional analytics."""

    def __init__(self, data: QueryData, analytics: DataAnalytics):
        self._data = data
        self._analytics = analytics

    def to_pandas(self) -> pd.DataFrame:
        """Convert the query data to a pandas DataFrame."""
        return pd.DataFrame(
            self._data.data, columns=self._data.columns
        ).convert_dtypes()

    @property
    def analytics(self) -> DataAnalytics:
        """Get the analytics data."""
        return self._analytics

    @property
    def data(self) -> QueryData:
        """Get the raw query data."""
        return self._data

    @staticmethod
    def from_jsonl_url(
        url: str,
    ) -> "QueryResponse":
        """Parse HTTP response chunks into QueryResponse."""
        analytics = {}

        response = requests.get(url, stream=True)
        response.raise_for_status()

        iterator = response.iter_lines(decode_unicode=True)
        columns = json.loads(next(iterator))
        data = [json.loads(line) for line in iterator]

        query_data = QueryData(columns=columns, data=data)
        return QueryResponse(data=query_data, analytics=DataAnalytics(analytics))


class JsonResponse(BaseModel):
    id: str
    status: Literal["queued", "running", "completed", "failed", "canceled"]
    url: Optional[str] = Field(default=None)


@lru_cache(maxsize=1)
def is_pyodide() -> bool:
    try:
        import sys

        import pyodide  # type: ignore # noqa: F401

        if sys.platform == "emscripten":
            return True
        return False
    except ImportError:
        return False


class Client:
    def __init__(
        self,
        api_key: Optional[str] = None,
        client_opts: Optional[ClientConfig] = None,
    ):
        self.__api_key = api_key if api_key else os.environ.get(OSO_API_KEY, "")
        if not self.__api_key and not is_pyodide():
            raise OsoError(
                "API key is required. Either set it in the environment variable OSO_API_KEY or pass it as an argument."
            )
        self.__base_url = DEFAULT_BASE_URL
        if client_opts and client_opts.base_url:
            self.__base_url = client_opts.base_url
            if not self.__base_url.endswith("/"):
                self.__base_url += "/"

        if HAS_OSO_SEMANTIC:
            self.semantic = create_registry(
                self.__base_url, self.__api_key, self.to_pandas
            )

    def _get_api_key(self) -> str:
        """We wrap the api key retreival here so that can lazily load the api
        key in a pyodide environment.

        This is in case the environment variable is not set immediately on
        startup (which can happen).
        """
        if self.__api_key:
            return self.__api_key
        key = os.environ.get(OSO_API_KEY)
        if not key:
            raise OsoError("API key is required.")
        self.__api_key = key
        return key

    def __query(
        self,
        query: str,
        input_dialect="trino",
        output_dialect="trino",
        include_analytics: bool = True,
    ) -> QueryResponse:
        # The following checks are only for providing better error messages as
        # the oso api does _not_ support multiple queries nor the use of
        # semicolons.
        if not query:
            raise OsoError("Query cannot be empty.")
        parsed_query = parse(query, dialect=input_dialect)
        if len(parsed_query) != 1:
            raise OsoError(
                "Only single queries are supported. Please provide a single SQL statement."
            )
        query_expression = parsed_query[0]
        assert query_expression is not None, "query could not be parsed"

        return self._make_sql_request(
            translated_query=query_expression.sql(dialect=output_dialect),
            include_analytics=include_analytics,
        )

    def _make_sql_request(
        self,
        translated_query: str,
        include_analytics: bool = True,
    ) -> QueryResponse:
        headers = {
            "Content-Type": "application/json",
        }
        headers["Authorization"] = f"Bearer {self._get_api_key()}"
        try:
            response = requests.post(
                f"{self.__base_url}async-sql",
                headers=headers,
                json={
                    "query": translated_query,
                },
            )
            response.raise_for_status()
            json_response = JsonResponse.model_validate(response.json())
            period = 1
            while json_response.status not in ("completed", "failed", "canceled"):
                response = requests.get(
                    f"{self.__base_url}async-sql",
                    headers=headers,
                    params={"id": json_response.id},
                )
                response.raise_for_status()
                json_response = JsonResponse.model_validate(response.json())
                time.sleep(period)
                period = min(period * 2, 5)
            if json_response.status != "completed" or json_response.url is None:
                raise requests.HTTPError(
                    f"Query failed with status: {json_response.status}",
                    response=response,
                )

            return QueryResponse.from_jsonl_url(json_response.url)
        except requests.HTTPError as e:
            if e.response.status_code == 402:
                error_message = e.response.json()["error"]
                raise InsufficientCreditError(error_message) from None
            raise OsoHTTPError(e, response=e.response) from None

    def to_pandas(self, query: str):
        query_response = self.__query(query)
        return query_response.to_pandas()

    def query(self, query: str, include_analytics: bool = True) -> QueryResponse:
        """Execute a SQL query and return the full response including analytics data."""
        return self.__query(query, include_analytics=include_analytics)

    def dbapi_connection(self):
        """Get a DBAPI 2.0 compatible connection.

        This is experimental and incomplete. It will error unless forced right now.
        """
        from .engine import PyosoDBApiConnection

        return PyosoDBApiConnection(self)



# ruff: noqa: F401

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("pyoso")
except PackageNotFoundError:
    __version__ = "unknown"

from .analytics import DataAnalytics
from .client import Client, ClientConfig, QueryResponse
from .exceptions import HTTPError, InsufficientCreditError, OsoError
